/**************************************************************
  (C) 1997 by Achim Kaiser

  You may distribute under the terms of the GNU General Public
  License as specified in the COPYING file.
**************************************************************/

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/xpm.h>
#include <X11/Xatom.h>
#include <X11/cursorfont.h>
#ifdef SUNOS
#include <signal.h>
#else
#include <sys/signal.h>
#endif
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "positiontext.h"
#include "iwc.h"
#include "gui.h"
#include "timer.h"
#include "iconmanager.h"
#include "global.h"
#include "testinstall.h"
#include "windowname.h"
#include "options.h"
#include "color.h"
#include "includepixmaps/icon.xpm"

void setsize(Display *dis,Window win,Window movewin,int ax,int ay,int sx,int sy) 
{
  int x=0,y=0;
  unsigned int width=0,height=0;
  unsigned int border,depth;
  int gax,gay;
  div_t p;
  XEvent ev;
  Window root; 
  XGetGeometry(dis,XDefaultRootWindow(dis),&root,&x,&y,&width,&height,&border,&depth);
  if ((ax+sx/2)<0 ||
      (ax+sx/2)>width ||
      (ay+sy/2)<0 ||
      (ay+sy/2)>height) {
    ax=ax+sx/2;
    ay=ay+sy/2;
    p=div(ax,width);
    if (ax<0)
      ax=ax-(p.quot-1)*width;
    else
      ax=p.rem;
    p=div(ay,height);
    if (ay<0)
      ay=ay-(p.quot-1)*height;
    else
      ay=p.rem;
    ax=ax-sx/2;
    ay=ay-sy/2;
  }
  do {} while (GetGreatestParent(win)==win);
  if (movewin!=0) {
    do {
      XNextEvent(dis,&ev);  
    } while(ev.type!=MapNotify);
    XMoveResizeWindow(dis,win,ax,ay,sx,sy+1);
  } else
    XMoveResizeWindow(dis,win,ax,ay,sx,sy);
  if (movewin!=0) {
    do {
      XNextEvent(dis,&ev);  
    } while(ev.type!=ConfigureNotify);
  } else
    system("");
  XFlush(dis);
  XSync(dis,False);
  XTranslateCoordinates(GetDisplay(),win,XDefaultRootWindow(GetDisplay()),
			0,0,&x,&y,&root);
  gax=ax-(x-ax);
  gay=ay-(y-ay);
  XMoveResizeWindow(dis,win,gax,gay,sx,sy);
  if (movewin!=0)
    XMoveResizeWindow(dis,movewin,0,0,sx,sy);
}

int main(int argc,char *argv[])  
{
  Display *dis;
  Window win,movewin,detailwin;
  Font font;
  Window *oldwin;
  GC gc;
  XEvent event;
  XSizeHints hint;
  int scr;
  unsigned long fg, bg;
  int done;
  int mapagain;
  int maindesktop;
  XWindowAttributes rootatr;
  XSetWindowAttributes xswa;
  unsigned long valuemask;
  char *windowname=NULL;
  char *iconname=NULL;
  char *pathname=NULL;
  char *backgroundname=NULL;
  char *positiona=NULL;
  char *positions=NULL;
  Pixmap backgroundpixmap;
  Pixmap wm_iconpix;
  static Atom           proto_atom= None, delete_atom= None;
  XpmAttributes attr;

  /*debugging*/
  if (fork()!=0)
    exit(0);
  /***********/
  initgui(argc,argv);
  chdir((char*)getenv("HOME"));
  dis=XOpenDisplay((char*)getenv("DISPLAY"));
  if (dis==NULL) {
     printf("DFM: Can't open display! Check DISPLAY variable!\n");
     exit(-1);
  }
  SetDisplay(dis);
  SetTimer(0);
  if (IsDesktop()==True) {
    if (argc==3 && strcmp(argv[1],"-setdir")==0)
      if (strcmp("./.dfmdesk/",argv[2])==0)
	MainWindowOptions(argc,argv);
      else
	WindowOptions(argc,argv);      
    if (argc==3 && strcmp(argv[1],"-setfile")==0)
      FileOptions(argc,argv);
    if (argc>3 && strcmp(argv[1],"-setfile")==0)
      MultipleFileOptions(argc,argv);
  }
  if (argc==3 && strcmp(argv[2],"-detail")==0)
    SetKindOfWindow(DETAIL);
  else if (argc==3 && strcmp(argv[2],"-structure")==0)
    SetKindOfWindow(STRUCTURE);
  else  
    SetKindOfWindow(NORMAL);
  SetCursors(XCreateFontCursor(dis,XC_top_left_arrow),XCreateFontCursor(dis,XC_watch));
  scr=DefaultScreen(dis);
  bg=WhitePixel(dis,scr);
  fg=BlackPixel(dis,scr);
  if (argc==1) {
    hint.x=0; hint.y=0;
    hint.width=2;hint.height=2;
    hint.flags=PPosition|PSize;
    if (IsDesktop()==False) {
      maindesktop=True;
      
      windowname=".dfmdesk --- dfm:~/.dfmdesk/";
      iconname=".dfmdesk";
      pathname="~/.dfmdesk/";
    } else { 
      printf("DFM is already running.\n");
      XCloseDisplay(dis);
      exit(1);
    }
  } else {
    if (IsDesktop()==False) {
      printf("wait for mainprogram\n");
      signal(SIGALRM,destroy);
      alarm(30);
      do {} while(IsDesktop()==False);
      signal(SIGALRM,updatetimer);
    }
    pathname=malloc(3+strlen(argv[1]));
    pathname[0]=0;
    if (strncmp(argv[1],(char*)getenv("HOME"),strlen((char*)getenv("HOME")))==0 &&
        argv[1][strlen((char*)getenv("HOME"))]=='/') {
      strcat(pathname,"~");
      strcat(pathname,argv[1]+strlen((char*)getenv("HOME")));
    } else {
      if (argv[1][0]!='.' && argv[1][1]!='/' && argv[1][0]!='/')
	strcat(pathname,"~/");
      strcat(pathname,argv[1]);
    }
    if (pathname[0]=='.')
      pathname[0]='~';
    if (pathname[strlen(pathname)-1]!='/')
      strcat(pathname,"/");
    if (strcmp(pathname,"~/.dfmdesk/")==0 && GetKindOfWindow()==NORMAL) {
      XCloseDisplay(dis);
      exit(0);
    }

    windowname=GenerateWindowName(pathname,GetKindOfWindow());

    iconname=GenerateIconName(pathname);

    oldwin=GetWindowNamed(windowname);
    if (oldwin!=NULL) {
      XGetWindowAttributes(dis,*oldwin,&rootatr);
      if (rootatr.map_state==IsUnmapped) {
	XLowerWindow(dis,*oldwin);
	XMapWindow(dis,*oldwin);
	GetFileInfoFromDesktop(windowname,&positiona,&backgroundname,
			       &positions);
	GetPositionFromText(positiona,&hint.x,&hint.y,NULL);
	GetPositionFromText(positions,&hint.width,&hint.height,NULL);
	setsize(dis,*oldwin,0,hint.x,hint.y,hint.width,hint.height);
      }
      XRaiseWindow(dis,*oldwin);
      XCloseDisplay(dis);
      exit(0);
    }
    maindesktop=False;
    GetFileInfoFromDesktop(windowname,&positiona,&backgroundname,
			   &positions);
    GetPositionFromText(positiona,&hint.x,&hint.y,NULL);
    GetPositionFromText(positions,&hint.width,&hint.height,NULL);
    hint.flags=PPosition|PSize;
    if (positiona==NULL) {
      hint.x=80; hint.y=80;
      hint.width=500;hint.height=300;
      hint.flags=/*PPosition|*/PSize;
    }
  }
  if (testandinstall(pathname)==False) {
    printf("DFM: Can`t open \"%s\" ?!\n",pathname);
    XCloseDisplay(dis);
    exit(0);
  }
  SetIsDesktop(maindesktop);
  SetPath(pathname);
  SetWindowName(windowname);

  if (maindesktop==True) {
    xswa.override_redirect=True;
    valuemask=CWOverrideRedirect;  
    win=XCreateWindow(dis,DefaultRootWindow(dis),
		      hint.x,hint.y,hint.width,hint.height,
		      0,0,InputOnly,CopyFromParent,valuemask,&xswa);
    movewin=XCreateWindow(dis,win,
		      0,0,hint.width,hint.height,
		      0,0,InputOnly,CopyFromParent,valuemask,&xswa);
  } else {
    win=XCreateSimpleWindow(dis,DefaultRootWindow(dis),
			    hint.x,hint.y,hint.width,hint.height,
			    0,fg,bg);
    movewin=XCreateSimpleWindow(dis,win,
				0,0,hint.width,hint.height,
				0,fg,bg);
  }
  XDefineCursor(dis,win,GetWaitCursor());
  SetWindow(win);
  SetMoveWindow(movewin);
  if (maindesktop==True)
    LoadFileInfoToDesktop();
  attr.valuemask=XpmExactColors|XpmCloseness;
  attr.exactColors=False;
  attr.closeness=65536;
  XpmCreatePixmapFromData(GetDisplay(),GetWindow(),icon_xpm,
				&wm_iconpix,NULL,&attr); 
  XSetStandardProperties(dis,win,windowname,iconname,wm_iconpix,argv,argc,&hint);
  XLowerWindow(dis,win);
  XMapWindow(dis,movewin);
  UpdateWindow();
  UpdateFontName();
  UpdateTimer();
  if (maindesktop==False) {
    XSelectInput(dis,movewin,ButtonReleaseMask|ButtonPressMask|PointerMotionMask);
    XSelectInput(dis,win,StructureNotifyMask);
  } else
    XSelectInput(dis,movewin,EnterWindowMask|ButtonReleaseMask|ButtonPressMask|PointerMotionMask);
  if (GetKindOfWindow()==DETAIL) {
    detailwin=XCreateSimpleWindow(dis,win,20,0,732,20,
				  0,fg,bg);
    gc=XCreateGC(GetDisplay(),detailwin,0,0);
    font=XLoadFont(GetDisplay(),"8x13");
    XSetFont(GetDisplay(),gc,font);
    bg=GetBestColor(0xcf3c,0xcf3c,0xcf3c);
    backgroundpixmap=XCreatePixmap(GetDisplay(),detailwin,
				   732,20,XDefaultDepth(GetDisplay(),
							XDefaultScreen(GetDisplay())));
    XSetForeground(GetDisplay(),gc,bg);
    XFillRectangle(GetDisplay(),backgroundpixmap,gc,0,0,732,20);
    XSetBackground(GetDisplay(),gc,bg);
    XSetForeground(GetDisplay(),gc,fg);
    XDrawImageString(GetDisplay(),backgroundpixmap,gc,0,12," Size",strlen(" Size"));
    XDrawLine(GetDisplay(),backgroundpixmap,gc,0,0,0,19);
    XDrawImageString(GetDisplay(),backgroundpixmap,gc,116,12,"Name",strlen("Name"));
    XDrawLine(GetDisplay(),backgroundpixmap,gc,92,0,92,19);
    XDrawImageString(GetDisplay(),backgroundpixmap,gc,304,12,"Owner",strlen("Owner"));
    XDrawLine(GetDisplay(),backgroundpixmap,gc,300,0,300,19);
    XDrawImageString(GetDisplay(),backgroundpixmap,gc,376,12,"Group",strlen("Group"));
    XDrawLine(GetDisplay(),backgroundpixmap,gc,372,0,372,19);
    XDrawImageString(GetDisplay(),backgroundpixmap,gc,448,12,"Time",strlen("Time"));
    XDrawLine(GetDisplay(),backgroundpixmap,gc,444,0,444,19);
    XDrawImageString(GetDisplay(),backgroundpixmap,gc,648,12,"Modeline",strlen("Modelines"));
    XDrawLine(GetDisplay(),backgroundpixmap,gc,644,0,644,19);    
    XDrawLine(GetDisplay(),backgroundpixmap,gc,731,0,731,19);    
    XDrawLine(GetDisplay(),backgroundpixmap,gc,0,19,731,19);    
    XDrawLine(GetDisplay(),backgroundpixmap,gc,0,0,731,0);    
    XSetWindowBackgroundPixmap(GetDisplay(),detailwin,backgroundpixmap);
    XFreePixmap(GetDisplay(),backgroundpixmap);
    XFreeGC(dis,gc);
    XMapRaised(dis,detailwin);
    SetDetailWindow(detailwin);
  }
  if (maindesktop==False) {
    XMapRaised(dis,win);
    if (positiona!=NULL)
      setsize(dis,win,movewin,hint.x,hint.y,hint.width,hint.height);
    proto_atom = XInternAtom(dis, "WM_PROTOCOLS", False);
    delete_atom = XInternAtom(dis, "WM_DELETE_WINDOW", False);
    if ((proto_atom != None) && (delete_atom != None))
      XChangeProperty(dis, win, proto_atom, XA_ATOM, 32,
		      PropModePrepend, (unsigned char *) &delete_atom, 1);
  }
  if (GetKindOfWindow()==STRUCTURE)
    MakeWindowTree();
  else
    MakeWindowIcons();
  if (IsSubDesktop()==True) {
    SetWindowSaveStatus(GetPath(),True);
    SetIconSaveStatus(GetPath(),True);
  }
  done=0;
  mapagain=0;
  XDefineCursor(dis,win,GetNormCursor());
  signal(SIGALRM,updatetimer);
  alarm(GetTimer());
  while (True||done==0) {
    if (mapagain==0) {
      signal(SIGALRM,updatetimer);
      XNextEvent(dis,&event);
      signal(SIGALRM,repeat);
    } else
      XNextEvent(dis,&event);  
    switch(event.type) {
    case ClientMessage:
      XDefineCursor(dis,win,GetWaitCursor());
      if ((maindesktop==False) && 
	    (event.xclient.message_type == proto_atom) &&
	    (event.xclient.window == win) && 
	    (event.xclient.data.l[0] == delete_atom)) {
        XGetWindowAttributes(dis,win,&rootatr);
        if (rootatr.map_state==IsUnmapped || GetWindowCache()==0) {
            XCloseDisplay(GetDisplay());
            exit(0);
        }
    	XUnmapWindow(dis,win);
	mapagain=1;
        signal(SIGALRM,destroy);
	alarm(GetWindowCache());
      } else {
	if (mapagain==0)
	  WorkIncomingCall(event);
      }
      XDefineCursor(dis,win,GetNormCursor());
      break;
    case MapNotify:
      signal(SIGALRM,updatetimer);
      alarm(GetTimer());
      mapagain=0;
      break;
    case EnterNotify:
      XLowerWindow(dis,win);
      LowerIcons();
      ManageAutofocus(&event);
      break;
    case Expose:
      XLowerWindow(dis,win);
      LowerIcons();
      break;
    case LeaveNotify:
      XSelectInput(dis,movewin,EnterWindowMask|ButtonReleaseMask|ButtonPressMask|PointerMotionMask);
      XLowerWindow(dis,win);  
      break;
    case ButtonPress:
      if (maindesktop==True && event.xbutton.button==2) {	
	XSelectInput(dis,movewin,LeaveWindowMask);
	XLowerWindow(dis,win);  
	break;
      }
    case MotionNotify:
      if (maindesktop==True)
        XLowerWindow(dis,win);  
    default:
      if (mapagain==0) {
	done=HandleTheEvent(event);
      }
      break;
    }
  }
  if (maindesktop==True) {
    /*SaveFileInfoFromDesktop();*/
    /*system("xtexit");*/
  }
  XCloseDisplay(dis);
  exit(0);
}
